#include <stdbool.h>
#include <stdlib.h>
#include <string.h>

#include "gfx/effect.h"
#include "gfx/renderer_api.h"
#include "runtime/buffer_manager.h"
#include "runtime/handles.h"
#include "runtime/mem_arena.h"

#include "common_render_graph.h"

#define MAX_COLOR_ATTACHMENTS_PER_PASS 8
#define MAX_SAMPLED_INPUTS_PER_PASS    8

typedef struct rt_render_target_build_info {
    const char *name;
    rt_pixel_format format;
    unsigned int width;
    unsigned int height;

    unsigned int samples;
    unsigned int layers;

    uint32_t first_usage;
    uint32_t last_usage;
} rt_render_target_build_info;

typedef struct rt_pass_build_info {
    const char *name;
    uint32_t flags;

    void *userdata;
    rt_execute_render_pass_fn *Execute;

    uint32_t color_attachments[MAX_COLOR_ATTACHMENTS_PER_PASS];
    rt_color color_attachment_clear_values[MAX_COLOR_ATTACHMENTS_PER_PASS];
    rt_pass_load_mode color_attachment_loads[MAX_COLOR_ATTACHMENTS_PER_PASS];
    rt_pass_write_mode color_attachment_writes[MAX_COLOR_ATTACHMENTS_PER_PASS];
    uint32_t color_attachment_count;

    uint32_t sampled_inputs[MAX_SAMPLED_INPUTS_PER_PASS];
    uint32_t sampled_input_count;

    uint32_t depth_stencil_attachment;
    rt_depth_stencil_value depth_stencil_clear_value;
    rt_pass_load_mode depth_stencil_load;
    rt_pass_write_mode depth_stencil_write;

    uint32_t *dependencies;
    uint32_t dependency_count;
} rt_pass_build_info;

typedef struct {
    uint32_t signaled_by;
    uint32_t waited_on_by;
} rt_sync_point_build_info;

typedef struct rt_render_graph_builder_obj {
    rt_arena arena;

    rt_render_target_build_info *render_targets;
    uint32_t render_target_count;
    uint32_t render_target_capacity;

    rt_pass_build_info *passes;
    uint32_t pass_count;
    uint32_t pass_capacity;

    rt_physical_render_target_info *phys_render_targets;
    uint32_t phys_render_target_count;

    rt_sync_point_build_info *sync_points;
    uint32_t sync_point_count;

    uint32_t backbuffer;

    rt_render_graph_builder_platform_callbacks platform_cbs;
} rt_render_graph_builder_obj;

/* ****************************************************************************
 *
 *                           BUILDER CODE
 *
 * ****************************************************************************/

static void AddRenderTarget(void *_obj, const rt_attachment_info *info) {
    rt_render_graph_builder_obj *obj = _obj;

    if (obj->render_target_count == obj->render_target_capacity) {
        uint32_t new_cap = obj->render_target_capacity ? 2 * obj->render_target_capacity : 32;
        rt_render_target_build_info *tmp =
            RT_ARENA_PUSH_ARRAY_ZERO(&obj->arena, rt_render_target_build_info, new_cap);
        if (obj->render_target_capacity)
            memcpy(tmp,
                   obj->render_targets,
                   sizeof(rt_render_target_build_info) * obj->render_target_capacity);
        obj->render_targets         = tmp;
        obj->render_target_capacity = new_cap;
    }

    char *name = rtArenaPush(&obj->arena, strlen(info->name) + 1);
    strcpy(name, info->name);
    obj->render_targets[obj->render_target_count].name        = name;
    obj->render_targets[obj->render_target_count].format      = info->format;
    obj->render_targets[obj->render_target_count].width       = info->width;
    obj->render_targets[obj->render_target_count].height      = info->height;
    obj->render_targets[obj->render_target_count].samples     = info->samples;
    obj->render_targets[obj->render_target_count].layers      = info->layers;
    obj->render_targets[obj->render_target_count].first_usage = 0;
    obj->render_targets[obj->render_target_count].last_usage  = 0;
    ++obj->render_target_count;
}

static void SetBackbuffer(void *_obj, const char *rt_name) {
    rt_render_graph_builder_obj *obj = _obj;
    for (uint32_t i = 0; i < obj->render_target_count; ++i) {
        if (strcmp(obj->render_targets[i].name, rt_name) == 0) {
            obj->backbuffer = i;
            return;
        }
    }
    rtLog("vk", "Tried to set backbuffer to unknown render target %s", rt_name);
}

static void AddRenderPass(void *_obj, const rt_pass_info *info) {
    rt_render_graph_builder_obj *obj = _obj;

    if (obj->pass_count == obj->pass_capacity) {
        uint32_t new_cap = obj->pass_capacity ? 2 * obj->pass_capacity : 32;
        rt_pass_build_info *tmp =
            RT_ARENA_PUSH_ARRAY_ZERO(&obj->arena, rt_pass_build_info, new_cap);
        if (obj->pass_capacity)
            memcpy(tmp, obj->passes, sizeof(rt_pass_build_info) * obj->pass_capacity);
        obj->passes        = tmp;
        obj->pass_capacity = new_cap;
    }

    char *name = rtArenaPush(&obj->arena, strlen(info->name) + 1);
    strcpy(name, info->name);
    obj->passes[obj->pass_count].name                     = name;
    obj->passes[obj->pass_count].flags                    = info->flags;
    obj->passes[obj->pass_count].color_attachment_count   = 0;
    obj->passes[obj->pass_count].sampled_input_count      = 0;
    obj->passes[obj->pass_count].depth_stencil_attachment = UINT_MAX;
    obj->passes[obj->pass_count].dependencies             = NULL;
    obj->passes[obj->pass_count].dependency_count         = 0;

    ++obj->pass_count;
}

static void AddColorOutput(void *_obj,
                           const char *pass_name,
                           const char *rt_name,
                           rt_pass_load_mode load,
                           rt_pass_write_mode write,
                           rt_color clear_color) {
    uint32_t rt_index = UINT_MAX;

    rt_render_graph_builder_obj *obj = _obj;
    for (uint32_t i = 0; i < obj->render_target_count; ++i) {
        if (strcmp(obj->render_targets[i].name, rt_name) == 0) {
            rt_index = i;
            break;
        }
    }
    if (rt_index == UINT_MAX) {
        rtLog("ren",
              "Tried to add unknown render target %s as color output to %s",
              rt_name,
              pass_name);
        return;
    }

    for (uint32_t i = 0; i < obj->pass_count; ++i) {
        if (strcmp(obj->passes[i].name, pass_name) == 0) {
            if (obj->passes[i].color_attachment_count == MAX_COLOR_ATTACHMENTS_PER_PASS) {
                rtLog("ren", "Too many color attachments in pass %s", pass_name);
            }
            obj->passes[i].color_attachment_clear_values[obj->passes[i].color_attachment_count] =
                clear_color;
            obj->passes[i].color_attachment_loads[obj->passes[i].color_attachment_count]  = load;
            obj->passes[i].color_attachment_writes[obj->passes[i].color_attachment_count] = write;
            obj->passes[i].color_attachments[obj->passes[i].color_attachment_count++] = rt_index;
            return;
        }
    }
    rtLog("ren",
          "Tried to add render target %s as color output to unknown render target %s",
          rt_name,
          pass_name);
}

static void AddSampledInput(void *_obj, const char *pass_name, const char *rt_name) {
    uint32_t rt_index = UINT_MAX;

    rt_render_graph_builder_obj *obj = _obj;
    for (uint32_t i = 0; i < obj->render_target_count; ++i) {
        if (strcmp(obj->render_targets[i].name, rt_name) == 0) {
            rt_index = i;
            break;
        }
    }
    if (rt_index == UINT_MAX) {
        rtLog("ren",
              "Tried to add unknown render target %s as color output to %s",
              rt_name,
              pass_name);
        return;
    }

    for (uint32_t i = 0; i < obj->pass_count; ++i) {
        if (strcmp(obj->passes[i].name, pass_name) == 0) {
            if (obj->passes[i].sampled_input_count == MAX_SAMPLED_INPUTS_PER_PASS) {
                rtLog("ren", "Too many sampled inputs in pass %s", pass_name);
            }
            obj->passes[i].sampled_inputs[obj->passes[i].sampled_input_count++] = rt_index;
            return;
        }
    }
    rtLog("ren",
          "Tried to add render target %s as sampled input to unknown render target %s",
          rt_name,
          pass_name);
}

static void SetDepthStencilAttachment(void *_obj,
                                      const char *pass_name,
                                      const char *rt_name,
                                      rt_pass_load_mode load,
                                      rt_pass_write_mode write,
                                      rt_depth_stencil_value clear_value) {
    uint32_t rt_index = UINT_MAX;

    rt_render_graph_builder_obj *obj = _obj;
    for (uint32_t i = 0; i < obj->render_target_count; ++i) {
        if (strcmp(obj->render_targets[i].name, rt_name) == 0) {
            rt_index = i;
            break;
        }
    }
    if (rt_index == UINT_MAX) {
        rtLog("ren",
              "Tried to add unknown render target %s as depth stencil attachment to %s",
              rt_name,
              pass_name);
        return;
    }

    for (uint32_t i = 0; i < obj->pass_count; ++i) {
        if (strcmp(obj->passes[i].name, pass_name) == 0) {
            obj->passes[i].depth_stencil_attachment  = rt_index;
            obj->passes[i].depth_stencil_clear_value = clear_value;
            obj->passes[i].depth_stencil_load        = load;
            obj->passes[i].depth_stencil_write       = write;
            return;
        }
    }
    rtLog("ren",
          "Tried to add render target %s as  depth stencil attachment to unknown render target %s",
          rt_name,
          pass_name);
}

static void BindRenderPass(void *_obj,
                           const char *pass_name,
                           rt_execute_render_pass_fn *execute_fn,
                           void *userdata) {
    rt_render_graph_builder_obj *obj = _obj;
    for (uint32_t i = 0; i < obj->pass_count; ++i) {
        if (strcmp(obj->passes[i].name, pass_name) == 0) {
            obj->passes[i].Execute  = execute_fn;
            obj->passes[i].userdata = userdata;
            return;
        }
    }
    rtLog("ren", "Tried to bind unknown render pass %s.", pass_name);
}

typedef struct {
    uint32_t added;
    uint32_t moved;
} rt_find_writers_result;

static rt_find_writers_result FindWriters(rt_render_graph_builder_obj *obj,
                                          uint32_t rt_index,
                                          uint32_t append_at,
                                          uint32_t *p_passes) {
    rt_find_writers_result res = {0, 0};
    for (uint32_t i = 0; i < obj->pass_count; ++i) {
        bool writes_rt = false;
        if (obj->passes[i].depth_stencil_attachment == rt_index) {
            writes_rt = true;
        } else {
            for (uint32_t j = 0; j < obj->passes[i].color_attachment_count; ++j) {
                if (obj->passes[i].color_attachments[j] == rt_index) {
                    writes_rt = true;
                }
            }
        }

        if (!writes_rt)
            continue;

        uint32_t lower_index = UINT32_MAX;
        for (uint32_t j = 0; j < append_at; ++j) {
            if (p_passes[j] == i) {
                lower_index = j;
                break;
            }
        }

        if (lower_index == UINT32_MAX) {
            p_passes[append_at++] = i;
            res.added++;
        } else {
            memmove(&p_passes[lower_index],
                    &p_passes[lower_index + 1],
                    (append_at - lower_index - 1) * sizeof(uint32_t));
            p_passes[append_at - 1] = i;
            res.moved++;
        }
    }
    return res;
}

static uint32_t CreateNaiveOrderRec(rt_render_graph_builder_obj *obj,
                                    uint32_t search_rt,
                                    uint32_t append_at,
                                    uint32_t *p_order) {
    rt_find_writers_result writers = FindWriters(obj, search_rt, append_at, p_order);
    uint32_t new_append            = append_at + writers.added;
    for (uint32_t i = 0; i < writers.moved; ++i) {
        uint32_t pass_idx              = p_order[append_at - writers.moved + i];
        const rt_pass_build_info *pass = &obj->passes[pass_idx];
        for (uint32_t j = 0; j < pass->sampled_input_count; ++j) {
            new_append = CreateNaiveOrderRec(obj, pass->sampled_inputs[j], new_append, p_order);
        }
    }
    for (uint32_t i = 0; i < writers.added; ++i) {
        uint32_t pass_idx              = p_order[append_at + i];
        const rt_pass_build_info *pass = &obj->passes[pass_idx];
        for (uint32_t j = 0; j < pass->sampled_input_count; ++j) {
            new_append = CreateNaiveOrderRec(obj, pass->sampled_inputs[j], new_append, p_order);
        }
    }
    return new_append;
}

static rt_result
CreateNaiveOrder(rt_render_graph_builder_obj *obj, uint32_t **p_order, uint32_t *p_count) {
    uint32_t *order = RT_ARENA_PUSH_ARRAY(&obj->arena, uint32_t, obj->pass_count);
    if (!order)
        return RT_OUT_OF_MEMORY;
    uint32_t count = CreateNaiveOrderRec(obj, obj->backbuffer, 0, order);

    /* Now the pass writing the backbuffer is first, we need to revert the order */
    for (uint32_t i = 0; i < count / 2; ++i) {
        uint32_t t           = order[i];
        order[i]             = order[count - i - 1];
        order[count - i - 1] = t;
    }
    *p_order = order;
    *p_count = count;
    return RT_SUCCESS;
}

static uint32_t *
ReorderPasses(rt_render_graph_builder_obj *obj, uint32_t pass_count, uint32_t *naive_order) {
    /* Our goal is to calculate a schedule that:
     * A) Does not break the dependency chain
     * B) Has the maximum amount of overlap, i.e. keeps the GPU busy.
     *    This means that if pass A depends on pass B, we want to have as much passes inbetween as
     *    possible, to reduce the likelyhood of stalls caused by B waiting for A to finish. */
    uint32_t *schedule = RT_ARENA_PUSH_ARRAY(&obj->arena, uint32_t, pass_count);
    if (!schedule)
        return NULL;
    uint32_t scheduled_count = 0;

    while (scheduled_count < pass_count) {
        /* The number of passes remaining in naive_order */
        uint32_t unscheduled_count = pass_count - scheduled_count;

        /* It is always valid to use the front */
        uint32_t selected_idx   = 0;
        uint32_t selected_score = 0;
        for (uint32_t i = 0; i < unscheduled_count; ++i) {
            /* Check if any dependency is not scheduled yet */
            uint32_t pass_idx              = naive_order[i];
            const rt_pass_build_info *pass = &obj->passes[pass_idx];
            uint32_t score                 = 0;
            bool is_valid                  = true;

            if (pass->dependency_count) {
                for (uint32_t j = 0; j < unscheduled_count; ++j) {
                    uint32_t pass2_idx = naive_order[j];
                    for (uint32_t k = 0; k < pass->dependency_count; ++k) {
                        if (pass->dependencies[k] == pass2_idx) {
                            is_valid = false;
                            break;
                        }
                    }
                    if (!is_valid)
                        break;
                }
                if (!is_valid)
                    continue;

                for (uint32_t j = 0; j < pass->dependency_count; ++j) {
                    for (uint32_t k = 0; k < scheduled_count; ++k) {
                        if (schedule[k] == pass->dependencies[j]) {
                            score += scheduled_count - k;
                            break;
                        }
                    }
                }

            } else {
                score = UINT32_MAX;
            }

            if (score > selected_score) {
                selected_score = score;
                selected_idx   = i;
            }
        }

        schedule[scheduled_count++] = naive_order[selected_idx];
        memmove(&naive_order[selected_idx],
                &naive_order[selected_idx + 1],
                (unscheduled_count - selected_idx - 1) * sizeof(uint32_t));
    }
    return schedule;
}

static rt_result DeterminePassDependencies(rt_render_graph_builder_obj *obj) {
    /* NOTE(Kevin): This has a lot of overlap with FindWriters, so maybe its possible to combine
     * the two */
    for (uint32_t pass_idx = 0; pass_idx < obj->pass_count; ++pass_idx) {
        rt_pass_build_info *pass     = &obj->passes[pass_idx];
        uint32_t dependency_capacity = pass->sampled_input_count;
        if (dependency_capacity) {
            pass->dependencies = RT_ARENA_PUSH_ARRAY(&obj->arena, uint32_t, dependency_capacity);
            if (!pass->dependencies)
                return RT_OUT_OF_MEMORY;
        }
        for (uint32_t input_idx = 0; input_idx < pass->sampled_input_count; ++input_idx) {
            uint32_t rt_index = pass->sampled_inputs[input_idx];
            for (uint32_t candidate_idx = 0; candidate_idx < obj->pass_count; ++candidate_idx) {
                const rt_pass_build_info *candidate = &obj->passes[candidate_idx];
                bool is_dependency                  = false;
                if (candidate->depth_stencil_attachment == rt_index)
                    is_dependency = true;
                for (uint32_t j = 0; j < candidate->color_attachment_count; ++j) {
                    if (candidate->color_attachments[j] == rt_index)
                        is_dependency = true;
                }

                if (!is_dependency)
                    continue;

                if (pass->dependency_count == dependency_capacity) {
                    /* The dependencies are still on top of the arena, so we can just grow that
                     * array */
                    if (!RT_ARENA_PUSH_ARRAY(&obj->arena, uint32_t, dependency_capacity))
                        return RT_OUT_OF_MEMORY;
                    dependency_capacity *= 2;
                }
                pass->dependencies[pass->dependency_count++] = candidate_idx;
            }
        }
    }
    return RT_SUCCESS;
}

static void DetermineRenderTargetUsage(rt_render_graph_builder_obj *obj,
                                       uint32_t pass_count,
                                       const uint32_t *schedule) {
    for (uint32_t rt_idx = 0; rt_idx < obj->render_target_count; ++rt_idx) {
        rt_render_target_build_info *rt = &obj->render_targets[rt_idx];
        rt->first_usage                 = UINT32_MAX;
        rt->last_usage                  = 0;
        for (uint32_t sched_idx = 0; sched_idx < pass_count; ++sched_idx) {
            uint32_t pass_idx              = schedule[sched_idx];
            const rt_pass_build_info *pass = &obj->passes[pass_idx];
            bool usage                     = pass->depth_stencil_attachment == rt_idx;
            if (!usage) {
                for (unsigned int i = 0; i < pass->color_attachment_count; ++i) {
                    if (pass->color_attachments[i] == rt_idx)
                        usage = true;
                }
            }
            if (!usage) {
                for (unsigned int i = 0; i < pass->sampled_input_count; ++i) {
                    if (pass->sampled_inputs[i] == rt_idx)
                        usage = true;
                }
            }
            if (usage) {
                if (sched_idx < rt->first_usage)
                    rt->first_usage = sched_idx;
                if (sched_idx > rt->last_usage)
                    rt->last_usage = sched_idx;
            }
        }
    }
}

static rt_result GreedyMergeRenderTargets(rt_render_graph_builder_obj *obj) {
    typedef struct {
        rt_physical_render_target_info info;
        int alive;
        int backbuffer;
        uint32_t first_usage;
        uint32_t last_usage;
    } merged_rts;

    merged_rts *merged = RT_ARENA_PUSH_ARRAY(&obj->arena, merged_rts, 2 * obj->render_target_count);
    if (!merged) {
        return RT_OUT_OF_MEMORY;
    }
    uint32_t candidate_count = obj->render_target_count;
    for (uint32_t i = 0; i < candidate_count; ++i) {
        merged[i].alive        = 1;
        merged[i].backbuffer   = (i == obj->backbuffer);
        merged[i].info.format  = obj->render_targets[i].format;
        merged[i].info.width   = obj->render_targets[i].width;
        merged[i].info.height  = obj->render_targets[i].height;
        merged[i].info.layers  = obj->render_targets[i].layers;
        merged[i].info.name    = obj->render_targets[i].name;
        merged[i].info.samples = obj->render_targets[i].samples;
        merged[i].first_usage  = obj->render_targets[i].first_usage;
        merged[i].last_usage   = obj->render_targets[i].last_usage;
    }

    uint32_t *rt_mapping =
        RT_ARENA_PUSH_ARRAY_ZERO(&obj->arena, uint32_t, obj->render_target_count);
    if (!rt_mapping)
        return RT_OUT_OF_MEMORY;
    for (uint32_t i = 0; i < obj->render_target_count; ++i)
        rt_mapping[i] = i;

    bool did_merge;
    do {
        did_merge = false;
        for (uint32_t first = 0; first < candidate_count - 1; ++first) {
            if (!merged[first].alive)
                continue;
            for (uint32_t second = first + 1; second < candidate_count; ++second) {
                if (!merged[second].alive)
                    continue;

                if (!((merged[first].last_usage < merged[second].first_usage) ||
                      (merged[second].last_usage < merged[first].first_usage)))
                    continue;

                if (!(merged[first].info.width == merged[second].info.width &&
                      merged[first].info.height == merged[second].info.height &&
                      merged[first].info.samples == merged[second].info.samples &&
                      merged[first].info.layers == merged[second].info.layers &&
                      merged[first].info.format == merged[second].info.format))
                    continue;

                merged[first].alive  = 0;
                merged[second].alive = 0;

                merged_rts combined = {
                    .alive       = 1,
                    .backbuffer  = merged[first].backbuffer || merged[second].backbuffer,
                    .first_usage = RT_MIN(merged[first].first_usage, merged[second].first_usage),
                    .last_usage  = RT_MAX(merged[first].last_usage, merged[second].last_usage),
                    .info        = merged[first].info,
                };
                char *combined_name = rtArenaPush(&obj->arena,
                                                  strlen(merged[first].info.name) +
                                                      strlen(merged[second].info.name) + 2);
                if (!combined_name)
                    return RT_OUT_OF_MEMORY;
                strcpy(combined_name, merged[first].info.name);
                strcat(combined_name, "+");
                strcat(combined_name, merged[second].info.name);
                combined.info.name = combined_name;

                /* Update mappings. If indes < render_target_count, than it refers to a
                 * logical render target. If not, it refers to a merged render target */
                if (first < obj->render_target_count) {
                    rt_mapping[first] = candidate_count;
                } else {
                    // Find mappings that refer to this index and update them
                    for (uint32_t i = 0; i < obj->render_target_count; ++i) {
                        if (rt_mapping[i] == first)
                            rt_mapping[i] = candidate_count;
                    }
                }
                if (second < obj->render_target_count) {
                    rt_mapping[second] = candidate_count;
                } else {
                    // Find mappings that refer to this index and update them
                    for (uint32_t i = 0; i < obj->render_target_count; ++i) {
                        if (rt_mapping[i] == second)
                            rt_mapping[i] = candidate_count;
                    }
                }

                RT_ASSERT(candidate_count < 2 * obj->render_target_count, "");
                merged[candidate_count++] = combined;
                did_merge                 = true;
                break;
            }
            if (did_merge)
                break;
        }
    } while (did_merge);

    uint32_t phys_count = 0;
    for (uint32_t i = 0; i < candidate_count; ++i) {
        if (merged[i].alive)
            ++phys_count;
    }
    obj->phys_render_targets =
        RT_ARENA_PUSH_ARRAY(&obj->arena, rt_physical_render_target_info, phys_count);
    if (!obj->phys_render_targets)
        return RT_OUT_OF_MEMORY;
    obj->phys_render_target_count = 0;
    for (uint32_t i = 0; i < candidate_count; ++i) {
        if (merged[i].alive) {
            uint32_t index = obj->phys_render_target_count;
            if (merged[i].backbuffer)
                obj->backbuffer = obj->phys_render_target_count;
            obj->phys_render_targets[obj->phys_render_target_count++] = merged[i].info;

            /* Update the mapping table */
            for (uint32_t j = 0; j < obj->render_target_count; ++j) {
                if (rt_mapping[j] == i)
                    rt_mapping[j] = index;
            }
        }
    }

    /* Update pass render target references */
    for (uint32_t i = 0; i < obj->pass_count; ++i) {
        if (obj->passes[i].depth_stencil_attachment < UINT_MAX)
            obj->passes[i].depth_stencil_attachment =
                rt_mapping[obj->passes[i].depth_stencil_attachment];
        for (uint32_t j = 0; j < obj->passes[i].color_attachment_count; ++j)
            obj->passes[i].color_attachments[j] = rt_mapping[obj->passes[i].color_attachments[j]];
        for (uint32_t j = 0; j < obj->passes[i].sampled_input_count; ++j)
            obj->passes[i].sampled_inputs[j] = rt_mapping[obj->passes[i].sampled_inputs[j]];
    }
    obj->backbuffer = rt_mapping[obj->backbuffer];

    return RT_SUCCESS;
}

static rt_result CreateSynchronizationPoints(rt_render_graph_builder_obj *obj) {
    RT_ASSERT(false, "Not implemented yet");
    return RT_UNKNOWN_ERROR;
}

static rt_render_graph *CreateRenderGraph(rt_render_graph_builder_obj *obj, const uint32_t *order) {

    size_t required_size = sizeof(rt_render_graph);
    required_size += obj->phys_render_target_count * sizeof(rt_render_target_handle);
    required_size += obj->sync_point_count * sizeof(rt_gpu_semaphore_handle);
    required_size += obj->pass_count * sizeof(rt_render_pass);

    size_t pass_attachment_size = 0;

    for (uint32_t i = 0; i < obj->pass_count; ++i) {
        required_size += strlen(obj->passes[i].name) + 1;
        pass_attachment_size += obj->passes[i].color_attachment_count *
                                (sizeof(rt_render_target_handle) + sizeof(rt_color) +
                                 sizeof(rt_pass_load_mode) + sizeof(rt_pass_write_mode));
        pass_attachment_size +=
            obj->passes[i].sampled_input_count * sizeof(rt_render_target_handle);
    }
    required_size += pass_attachment_size;

    rt_render_graph *graph = rtAllocBuffer(required_size);
    if (!graph)
        return NULL;
    memset(graph, 0, required_size);
    graph->render_targets = (rt_render_target_handle *)(graph + 1);
    graph->semaphores =
        (rt_gpu_semaphore_handle *)(graph->render_targets + obj->phys_render_target_count);
    graph->passes            = (rt_render_pass *)(graph->semaphores + obj->sync_point_count);
    char *attachment_storage = (char *)(graph->passes + obj->pass_count);
    char *names              = attachment_storage + pass_attachment_size;
    char *next_name          = names;

    graph->render_target_count = obj->phys_render_target_count;
    graph->semaphore_count     = obj->sync_point_count;
    graph->pass_count          = obj->pass_count;

    for (uint32_t i = 0; i < obj->phys_render_target_count; ++i) {
        graph->render_targets[i] =
            obj->platform_cbs.CreateRenderTarget(&obj->phys_render_targets[i]);
    }

    for (uint32_t i = 0; i < obj->sync_point_count; ++i) {
        // TODO
        RT_NOT_IMPLEMENTED;
    }

    for (uint32_t i = 0; i < obj->pass_count; ++i) {
        uint32_t passidx              = order[i];
        size_t namelen                = strlen(obj->passes[passidx].name);
        graph->passes[i].Execute      = RT_VERIFY(obj->passes[passidx].Execute);
        graph->passes[i].user_data    = obj->passes[passidx].userdata;
        graph->passes[i].flags        = obj->passes[passidx].flags;
        graph->passes[i].id           = rtCalculateRenderPassID(obj->passes[passidx].name, namelen);
        graph->passes[i].first_signal = 0;
        graph->passes[i].signal_count = 0;
        graph->passes[i].first_wait   = 0;
        graph->passes[i].wait_count   = 0;
        graph->passes[i].execution_level = i;

        graph->passes[i].depth_stencil =
            (obj->passes[i].depth_stencil_attachment != UINT_MAX)
                ? graph->render_targets[obj->passes[i].depth_stencil_attachment]
                : (rt_render_target_handle)RT_INVALID_HANDLE;
        graph->passes[i].depth_stencil_clear_value = obj->passes[i].depth_stencil_clear_value;
        graph->passes[i].depth_stencil_load        = obj->passes[i].depth_stencil_load;
        graph->passes[i].depth_stencil_write       = obj->passes[i].depth_stencil_write;

        graph->passes[i].color_output_count = obj->passes[i].color_attachment_count;
        if (graph->passes[i].color_output_count) {
            graph->passes[i].color_outputs = (rt_render_target_handle *)attachment_storage;
            attachment_storage +=
                sizeof(rt_render_target_handle) * graph->passes[i].color_output_count;
            graph->passes[i].color_clear_values = (rt_color *)attachment_storage;
            attachment_storage += sizeof(rt_color) * graph->passes[i].color_output_count;
            graph->passes[i].color_loads = (rt_pass_load_mode *)attachment_storage;
            attachment_storage += sizeof(rt_pass_load_mode) * graph->passes[i].color_output_count;
            graph->passes[i].color_writes = (rt_pass_write_mode *)attachment_storage;
            attachment_storage += sizeof(rt_pass_write_mode) * graph->passes[i].color_output_count;

            for (uint32_t j = 0; j < graph->passes[i].color_output_count; ++j) {
                graph->passes[i].color_outputs[j] =
                    graph->render_targets[obj->passes[i].color_attachments[j]];
                graph->passes[i].color_clear_values[j] =
                    obj->passes[i].color_attachment_clear_values[j];
                graph->passes[i].color_loads[j]  = obj->passes[i].color_attachment_loads[j];
                graph->passes[i].color_writes[j] = obj->passes[i].color_attachment_writes[j];
            }
        }

        graph->passes[i].sampled_input_count = obj->passes[i].sampled_input_count;
        if (graph->passes[i].sampled_input_count) {
            graph->passes[i].sampled_inputs = (rt_render_target_handle *)attachment_storage;
            attachment_storage +=
                sizeof(rt_render_target_handle) * graph->passes[i].sampled_input_count;

            for (uint32_t j = 0; j < graph->passes[i].sampled_input_count; ++j) {
                graph->passes[i].sampled_inputs[j] =
                    graph->render_targets[obj->passes[i].sampled_inputs[j]];
            }
        }

        graph->passes[i].name = next_name;
        next_name += namelen + 1;
        memcpy((char *)graph->passes[i].name, obj->passes[passidx].name, namelen + 1);
    }

    graph->backbuffer_index = obj->backbuffer;

    return graph;
}

static rt_result Build(void *_obj, rt_render_graph **p_graph) {
    rt_render_graph_builder_obj *obj = _obj;
    uint32_t *naive_order;
    uint32_t pass_count;
    rt_result res = CreateNaiveOrder(obj, &naive_order, &pass_count);
    if (res != RT_SUCCESS)
        return res;

    res = DeterminePassDependencies(obj);
    if (res != RT_SUCCESS)
        return res;

    uint32_t *optimized_order = ReorderPasses(obj, pass_count, naive_order);
    if (!optimized_order)
        return RT_OUT_OF_MEMORY;

    DetermineRenderTargetUsage(obj, pass_count, optimized_order);
    res = GreedyMergeRenderTargets(obj);
    if (res != RT_SUCCESS)
        return res;

    if (obj->platform_cbs.RequireExplicitSynchronization()) {
        res = CreateSynchronizationPoints(obj);
        if (res != RT_SUCCESS)
            return res;
    } else {
        obj->sync_point_count = 0;
    }

    *p_graph = CreateRenderGraph(obj, optimized_order);
    return *p_graph ? RT_SUCCESS : RT_UNKNOWN_ERROR;
}

rt_render_graph_builder
rtCreateRenderGraphBuilder(const rt_render_graph_builder_platform_callbacks *platform_cbs) {
    // TODO: This is totally temporary. We do NOT want malloc anywhere in non-init functions
    rt_render_graph_builder_obj *obj = malloc(sizeof(*obj));
    RT_ASSERT(obj, "Failed to allocate the builder object.");
    memset(obj, 0, sizeof(*obj));
    rt_create_arena_result arena_res = rtCreateArena(NULL, RT_MB(16));
    RT_ASSERT(arena_res.ok, "");
    obj->arena        = arena_res.arena;
    obj->platform_cbs = *platform_cbs;

    return (rt_render_graph_builder){
        .obj                       = obj,
        .AddRenderTarget           = AddRenderTarget,
        .SetBackbuffer             = SetBackbuffer,
        .AddRenderPass             = AddRenderPass,
        .AddColorOutput            = AddColorOutput,
        .AddSampledInput           = AddSampledInput,
        .SetDepthStencilAttachment = SetDepthStencilAttachment,
        .BindRenderPass            = BindRenderPass,
        .Build                     = Build,
    };
}

void rtDestroyRenderGraphBuilder(rt_render_graph_builder *builder) {
    rt_render_graph_builder_obj *obj = builder->obj;
    rtReleaseArena(&obj->arena);
    free(obj);
    memset(builder, 0, sizeof(*builder));
}
