#ifndef RT_DX11_DEVICE_OBJECTS_HPP
#define RT_DX11_DEVICE_OBJECTS_HPP

// Types containing various api objects
#include <stdint.h>
#include <d3d11.h>
#include <d3d11_1.h>

#include "runtime/runtime.h"

struct rt_render_target {
    // Only one of these should be valid
    ID3D11RenderTargetView *rtv;
    ID3D11DepthStencilView *dsv;

    ID3D11Texture2D *texture;

    rt_pixel_format format;

    uint32_t version;
    rt_render_target *next_free;

    RT_INLINE bool HasStencilComponent() const {
        return format == RT_PIXEL_FORMAT_DEPTH24_STENCIL8;
    }

    RT_INLINE bool IsColorRenderTarget() const {
        RT_ASSERT(!(rtv != nullptr && dsv != nullptr),
                  "A render target should not contain a render target and a depth stencil view");
        return rtv != nullptr;
    }

    RT_INLINE bool IsDepthStencilTarget() const {
        RT_ASSERT(!(rtv != nullptr && dsv != nullptr),
                  "A render target should not contain a render target and a depth stencil view");
        return dsv != nullptr;
    }
};

struct rt_command_buffer {
    // Only created once and then re-used.
    ID3D11DeviceContext1 *context;
    ID3DUserDefinedAnnotation *annotation;

    uint32_t version;
    rt_command_buffer *next_free;
};

struct rt_buffer {
    ID3D11Buffer *buffer;
    rt_buffer_type type;
    rt_buffer_usage usage;

    uint32_t version;
    rt_buffer *next_free;
};

struct rt_pipeline {
    ID3D11InputLayout *input_layout;
    ID3D11VertexShader *vertex_shader;
    ID3D11PixelShader *pixel_shader;

    ID3D11ComputeShader *compute_shader;


    rt_pipeline *next_free;
    uint32_t version;

    RT_INLINE bool IsComputePipeline() const {
        RT_ASSERT(!(compute_shader && (vertex_shader || pixel_shader)),
                  "A pipeline should contain either a compute shader or graphics shaders.");
        return compute_shader != nullptr;
    }
};

struct rt_render_target_create_info {
    rt_pixel_format format;
    uint32_t width;
    uint32_t height;
    const char *name;
};

rt_render_target_handle rtCreateRenderTarget(const rt_render_target_create_info &info);
void rtDestroyRenderTarget(rt_render_target_handle handle);

rt_result rtAllocCommandBuffers(uint32_t count, rt_command_buffer_handle *p_handles);
rt_result rtSubmitCommandBuffers(uint32_t count, const rt_command_buffer_handle *handles);

rt_render_target *rtGetRenderTarget(rt_render_target_handle handle);
rt_command_buffer *rtGetCommandBuffer(rt_command_buffer_handle handle);
rt_buffer *rtGetBuffer(rt_buffer_handle handle);
rt_pipeline *rtGetPipeline(rt_pipeline_handle handle);

#endif