#include <d3d11.h>
#include <d3d11_1.h>

#include "gfx/renderer_api.h"
#include "runtime/config.h"
#include "runtime/handles.h"
#include "runtime/threading_helpers.hpp"

#include "device_objects.hpp"
#include "gpu.hpp"

RT_CVAR_I(rt_Dx11MaxRenderTargets,
          "Maximum number of simultaneously existing render targets. Default: 128",
          128);

static rt_render_target *_render_targets;
static rt_render_target *_first_free;
static rt_mutex *_lock;

rt_result InitRenderTargetManagement() {
    _render_targets = reinterpret_cast<rt_render_target *>(
        calloc((size_t)rt_Dx11MaxRenderTargets.i, sizeof(rt_render_target)));
    if (!_render_targets) {
        return RT_OUT_OF_MEMORY;
    }

    _lock = rtCreateMutex();
    if (!_lock) {
        free(_render_targets);
        return RT_UNKNOWN_ERROR;
    }

    _render_targets[1].rtv     = g_gpu.swap_chain.rtv.Get();
    _render_targets[1].format  = RT_PIXEL_FORMAT_B8G8R8A8_SRGB;
    _render_targets[1].version = 1;

    _first_free = _render_targets + 2;
    for (int i = 0; i < rt_Dx11MaxRenderTargets.i; ++i) {
        _render_targets[i].next_free = &_render_targets[i + 1];
    }
    return RT_SUCCESS;
}

void ShutdownRenderTargetManagement() {
    // Swapchain rtv in slot 1 will be released elsewhere
    for (int i = 2; i < rt_Dx11MaxRenderTargets.i; ++i) {
        if (_render_targets[i].rtv)
            _render_targets[i].rtv->Release();
        if (_render_targets[i].dsv)
            _render_targets[i].dsv->Release();
        if (_render_targets[i].texture)
            _render_targets[i].texture->Release();
    }
    free(_render_targets);
    rtDestroyMutex(_lock);
}

rt_render_target_handle rtCreateRenderTarget(const rt_render_target_create_info &info) {
    rt_render_target *slot = nullptr;
    {
        auto lock_guard = rtAutoLock(_lock);
        slot            = _first_free;
        _first_free     = slot->next_free;
    }

    if (!slot) {
        rtLog("dx11",
              "Could not create a new render target, because all available slots are currently in "
              "use.");
        return RT_INVALID_HANDLE;
    }

    slot->format = info.format;

    uint32_t swapchain_width = 0, swapchain_height = 0;
    if (info.width == RT_RENDER_TARGET_SIZE_SWAPCHAIN ||
        info.height == RT_RENDER_TARGET_SIZE_SWAPCHAIN) {

        DXGI_SWAP_CHAIN_DESC desc;
        g_gpu.swap_chain.swap_chain->GetDesc(&desc);
        swapchain_width  = desc.BufferDesc.Width;
        swapchain_height = desc.BufferDesc.Height;
    }

    if (!rtIsDepthFormat(info.format)) {
        D3D11_TEXTURE2D_DESC tex_desc = {};
        tex_desc.Width =
            (info.width != RT_RENDER_TARGET_SIZE_SWAPCHAIN) ? info.width : swapchain_width;
        tex_desc.Height =
            (info.height != RT_RENDER_TARGET_SIZE_SWAPCHAIN) ? info.height : swapchain_height;
        tex_desc.MipLevels          = 1;
        tex_desc.ArraySize          = 1;
        tex_desc.Format             = rtConvertPixelFormat(info.format);
        tex_desc.SampleDesc.Count   = 1;
        tex_desc.SampleDesc.Quality = 0;
        tex_desc.Usage              = D3D11_USAGE_DEFAULT; // read and write
        tex_desc.BindFlags          = D3D11_BIND_RENDER_TARGET | D3D11_BIND_SHADER_RESOURCE;
        tex_desc.CPUAccessFlags     = 0; // none
        tex_desc.MiscFlags          = 0;
        if (FAILED(g_gpu.device->CreateTexture2D(&tex_desc, nullptr, &slot->texture))) {
            rtLog("dx11", "Failed to create backing texture for render target %s", info.name);
            auto lg         = rtAutoLock(_lock);
            slot->next_free = _first_free;
            _first_free     = slot;
            return RT_INVALID_HANDLE;
        }

        D3D11_RENDER_TARGET_VIEW_DESC rtv_desc = {};
        rtv_desc.Format                        = rtConvertPixelFormat(info.format);
        rtv_desc.ViewDimension                 = D3D11_RTV_DIMENSION_TEXTURE2D;
        rtv_desc.Texture2D.MipSlice            = 0;
        if (FAILED(g_gpu.device->CreateRenderTargetView(slot->texture, &rtv_desc, &slot->rtv))) {
            slot->texture->Release();
            rtLog("dx11",
                  "Failed to create the render target view for render target %s",
                  info.name);
            auto lg         = rtAutoLock(_lock);
            slot->next_free = _first_free;
            _first_free     = slot;
            return RT_INVALID_HANDLE;
        }

        slot->version  = (slot->version + 1) % RT_RENDER_BACKEND_HANDLE_MAX_VERSION;
        uint32_t index = static_cast<uint32_t>(slot - _render_targets);
        return {.version = slot->version, .index = index};
    } else {
        D3D11_TEXTURE2D_DESC tex_desc = {};
        tex_desc.Width =
            (info.width != RT_RENDER_TARGET_SIZE_SWAPCHAIN) ? info.width : swapchain_width;
        tex_desc.Height =
            (info.height != RT_RENDER_TARGET_SIZE_SWAPCHAIN) ? info.height : swapchain_height;
        tex_desc.MipLevels          = 1;
        tex_desc.ArraySize          = 1;
        tex_desc.Format             = rtConvertPixelFormat(info.format);
        tex_desc.SampleDesc.Count   = 1;
        tex_desc.SampleDesc.Quality = 0;
        tex_desc.Usage              = D3D11_USAGE_DEFAULT; // read and write
        tex_desc.BindFlags          = D3D11_BIND_DEPTH_STENCIL | D3D11_BIND_SHADER_RESOURCE;
        tex_desc.CPUAccessFlags     = 0; // none
        tex_desc.MiscFlags          = 0;
        if (FAILED(g_gpu.device->CreateTexture2D(&tex_desc, nullptr, &slot->texture))) {
            rtLog("dx11", "Failed to create backing texture for render target %s", info.name);
            auto lg         = rtAutoLock(_lock);
            slot->next_free = _first_free;
            _first_free     = slot;
            return RT_INVALID_HANDLE;
        }

        D3D11_DEPTH_STENCIL_VIEW_DESC dsv_desc = {};
        dsv_desc.Format                        = rtConvertPixelFormat(info.format);
        dsv_desc.Flags                         = 0;
        dsv_desc.ViewDimension                 = D3D11_DSV_DIMENSION_TEXTURE2D;
        dsv_desc.Texture2D.MipSlice            = 0;
        if (FAILED(g_gpu.device->CreateDepthStencilView(slot->texture, &dsv_desc, &slot->dsv))) {
            slot->texture->Release();
            rtLog("dx11",
                  "Failed to create the depth stencil view for render target %s",
                  info.name);
            auto lg         = rtAutoLock(_lock);
            slot->next_free = _first_free;
            _first_free     = slot;
            return RT_INVALID_HANDLE;
        }

        slot->version  = (slot->version + 1) % RT_RENDER_BACKEND_HANDLE_MAX_VERSION;
        uint32_t index = static_cast<uint32_t>(slot - _render_targets);
        return {.version = slot->version, .index = index};
    }
}

void rtDestroyRenderTarget(rt_render_target_handle handle) {
    RT_UNUSED(handle);
}

rt_render_target *rtGetRenderTarget(rt_render_target_handle handle) {
    if (!RT_IS_HANDLE_VALID(handle) || (int)handle.index >= rt_Dx11MaxRenderTargets.i)
        return nullptr;
    auto lg = rtAutoLock(_lock);
    if (_render_targets[handle.index].version != handle.version)
        return nullptr;
    return &_render_targets[handle.index];
}
